/* $Id: Rubik.c,v 1.7 2003/01/28 09:11:15 d Exp $ */
/* David Leonard, 2003. Public domain. */

/*
 * A widget to display a Rubik's cube.
 * XXX Only three faces of the cube are ever shown!!
 */

#include <stdio.h>
#include <math.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "RubikP.h"
#include "cube.h"

/*
 * Class Methods
 */
static void	RubikInitialize(Widget, Widget, ArgList, Cardinal *);
static void	RubikExpose(Widget, XEvent *, Region);
/*static Boolean	RubikSetValues(Widget, Widget, Widget, ArgList, Cardinal *);*/
static void	RubikDestroy(Widget);
static void	RubikResize(Widget);
static void	RubikRealize(Widget w, XtValueMask *value_mask,
			XSetWindowAttributes *attributes);

/*
 * Actions
 */
static void	Turn(Widget, XEvent *, String *, Cardinal *);
static void	Roll(Widget, XEvent *, String *, Cardinal *);
static void	Reset(Widget, XEvent *, String *, Cardinal *);
static void	Undo(Widget, XEvent *, String *, Cardinal *);

/* 
 * Prototypes
 */
static void	drawcube(Widget w, int fast);
static void	redisplay(Widget w);
static int	parsemotion(Widget, String *params, Cardinal num_params,
			int *dirp, int *sidep);
static void	perspective(RubikWidget, double, double, double, XPoint *);
static void	perspectivef(RubikWidget, double, double, double, double *, double *);
static void	compute_cube_offset(RubikWidget rw);

/*
 * Resources
 */
#define offset(field)	XtOffsetOf(RubikRec, rubik.field)
static XtResource resources[] = {
/*	{ name, class, type,
	  size, offset,
	  default_type, default_addr }, */
	{ XtNtop, XtCTop, XtRPixel,
	  sizeof (Pixel), offset(top),
	  XtRString, XtDefaultBackground },
	{ XtNfront, XtCFront, XtRPixel,
	  sizeof (Pixel), offset(front),
	  XtRString, XtDefaultBackground },
	{ XtNright, XtCRight, XtRPixel,
	  sizeof (Pixel), offset(right),
	  XtRString, XtDefaultBackground },
	{ XtNbottom, XtCBottom, XtRPixel,
	  sizeof (Pixel), offset(bottom),
	  XtRString, XtDefaultBackground },
	{ XtNback, XtCBack, XtRPixel,
	  sizeof (Pixel), offset(back),
	  XtRString, XtDefaultBackground },
	{ XtNleft, XtCLeft, XtRPixel,
	  sizeof (Pixel), offset(left),
	  XtRString, XtDefaultBackground },
	{ XtNfocalDist, XtCFocalDist, XtRFloat,
	  sizeof (float), offset(focal_dist),
	  XtRString, "30" },
	{ XtNelevation, XtCElevation, XtRFloat,
	  sizeof (float), offset(elevation),
	  XtRString, "-10" },
	{ XtNedgeThickness, XtCEdgeThickness, XtRDimension,
	  sizeof (Dimension), offset(edge_thickness),
	  XtRString, "3" },
};
#undef offset

static XtActionsRec actions[] = {
/*	{ name,		procedure }, */
	{ "Turn",	Turn },
	{ "Roll",	Roll },
	{ "Reset",	Reset },
	{ "Undo",	Undo },
};

/*
 * Class record (methods and properties)
 */
RubikClassRec rubikClassRec = {
    /* core */
    {
	(WidgetClass)&widgetClassRec,		/* superclass */
	"Rubik",				/* class_name */
	sizeof (RubikRec),			/* widget_size */
	NULL,					/* class_initialize */
	NULL,					/* class_part_initialize */
	False,					/* class_inited */
	RubikInitialize,			/* initialize */
	NULL,					/* initialize_hook */
	RubikRealize,				/* realize */
	actions,				/* actions */
	XtNumber(actions),			/* num_actions */
	resources,				/* resources */
	XtNumber(resources),			/* num_resources */
	NULLQUARK,				/* xrm_class */
	True,					/* compress_motion */
	True,					/* compress_exposure */
	True,					/* compress_enterleave */
	False,					/* visible_interest */
	RubikDestroy,				/* destroy */
	RubikResize /*XtInheritResize*/,	/* resize */
	RubikExpose,				/* expose */
	NULL /*RubikSetValues */,		/* set_values */
	NULL,					/* set_values_hook */
	XtInheritSetValuesAlmost,		/* set_values_almost */
	NULL,					/* get_values_hook */
	XtInheritAcceptFocus,			/* accept_focus */
	XtVersion,				/* version */
	NULL,					/* callback_private */
	NULL,					/* tm_table */
	XtInheritQueryGeometry,			/* query_geometry */
	XtInheritDisplayAccelerator,		/* display_accelerator */
	NULL,					/* extension */
    },
    /* rubik */
    {
	NULL,					/* extension */
    }
};
WidgetClass rubikWidgetClass = (WidgetClass)&rubikClassRec;

/*
 * Class methods implementation
 */
static void
RubikInitialize(Widget request, Widget w, ArgList args, Cardinal *num_args)
{
	/* RubikWidget rw = (RubikWidget)w; */
}

static void
RubikRealize(Widget w, XtValueMask *value_mask, 
	XSetWindowAttributes *attributes)
{
	RubikWidget rw = (RubikWidget)w;
	XtRealizeProc realize;

	if ((realize = XtSuperclass(w)->core_class.realize))
		(*realize)(w, value_mask, attributes);

	compute_cube_offset(rw);
	cube_reset();
	cube_checkpoint();
	rw->rubik.undo_action = NULL;
}

static void
RubikExpose(Widget w, XEvent *e, Region r)
{
	drawcube(w, 0);
}

static void
drawcube(Widget w, int fast)
{
	RubikWidget rw = (RubikWidget)w;
	Display *display = XtDisplay(w);
	Drawable d = XtWindow(w);
	XPoint p[5];
	int i, x, y, z;
	GC gc, cgc, fgc[6];
	Pixel facepixel[6] = {
		rw->rubik.top,
		rw->rubik.front,
		rw->rubik.right,
		rw->rubik.back,
		rw->rubik.bottom,
		rw->rubik.left,
	};
	XGCValues values;

	gc = XtGetGC(w, 0, NULL);	/* get normal bg/fg GC */
	for (i = 0; i < 6; i++) {
		fgc[i] = XCreateGC(display, d, 0, NULL);
		XCopyGC(display, gc, ~(GCForeground), fgc[i]);
		values.foreground = facepixel[i];
		XChangeGC(display, fgc[i], GCForeground, &values);
	}

	/* Draw filled faces */

	z = 0;
	for (x = 0; x < 3; x++)
	    for (y = 0; y < 3; y++)
	      if (!fast || cube_changed(0, 2-x, y)) {
		perspective(rw, (x+0), (y+0), z, &p[0]);
		perspective(rw, (x+1), (y+0), z, &p[1]);
		perspective(rw, (x+1), (y+1), z, &p[2]);
		perspective(rw, (x+0), (y+1), z, &p[3]);
		memcpy(&p[4], &p[0], sizeof p[4]);
		cgc = fgc[cube_get(0, 2-x, y)];
		XFillPolygon(display, d, cgc, p, 5, Convex, CoordModeOrigin);
	      }

	x = 0;
	for (z = 0; z < 3; z++)
	    for (y = 0; y < 3; y++)
	      if (!fast || cube_changed(1, 2-y, z)) {
		perspective(rw, x, (y+0), (z+0), &p[0]);
		perspective(rw, x, (y+0), (z+1), &p[1]);
		perspective(rw, x, (y+1), (z+1), &p[2]);
		perspective(rw, x, (y+1), (z+0), &p[3]);
		memcpy(&p[4], &p[0], sizeof p[4]);
		cgc = fgc[cube_get(1, 2-y, z)];
		XFillPolygon(display, d, cgc, p, 5, Convex, CoordModeOrigin);
	      }

	y = 0;
	for (x = 0; x < 3; x++)
	    for (z = 0; z < 3; z++) 
	      if (!fast || cube_changed(2, 2-z, x)) {
		perspective(rw, (x+0), y, (z+0), &p[0]);
		perspective(rw, (x+1), y, (z+0), &p[1]);
		perspective(rw, (x+1), y, (z+1), &p[2]);
		perspective(rw, (x+0), y, (z+1), &p[3]);
		memcpy(&p[4], &p[0], sizeof p[4]);
		cgc = fgc[cube_get(2, 2-z, x)];
		XFillPolygon(display, d, cgc, p, 5, Convex, CoordModeOrigin);
	     }

	for (i = 0; i < 6; i++)
		XFreeGC(display, fgc[i]);

	cube_checkpoint();

	/* Draw outlines */

	cgc = XCreateGC(display, d, 0, NULL);
	XCopyGC(display, gc, ~(GCLineWidth|GCCapStyle|GCJoinStyle), cgc);
	values.line_width = rw->rubik.edge_thickness;
	values.cap_style = CapRound;
	values.join_style = JoinRound;
	XChangeGC(display, cgc, GCLineWidth|GCCapStyle|GCJoinStyle, &values);

	for (x = 0; x <= 3; x++) {
	    perspective(rw, x, 3, 0, &p[0]);
	    perspective(rw, x, 0, 0, &p[1]);
	    perspective(rw, x, 0, 3, &p[2]);
	    XDrawLines(display, d, cgc, p, 3, CoordModeOrigin);
	}

        perspective(rw, 3, 0, 0, &p[0]);
        perspective(rw, 0, 0, 0, &p[1]);
        XDrawLines(display, d, cgc, p, 2, CoordModeOrigin);

	for (y = 1; y <= 3; y++) {
	    perspective(rw, 3, y, 0, &p[0]);
	    perspective(rw, 0, y, 0, &p[1]);
	    perspective(rw, 0, y, 3, &p[2]);
	    XDrawLines(display, d, cgc, p, 3, CoordModeOrigin);
	}

	for (z = 1; z <= 3; z++) {
	    perspective(rw, 3, 0, z, &p[0]);
	    perspective(rw, 0, 0, z, &p[1]);
	    perspective(rw, 0, 3, z, &p[2]);
	    XDrawLines(display, d, cgc, p, 3, CoordModeOrigin);
	}

	XFreeGC(display, cgc);
	XtReleaseGC(w, gc);
}

#if 0
static Boolean
RubikSetValues(Widget current, Widget request, Widget w, 
	ArgList args, Cardinal *num_args)
{
/*
	RubikWidget rcurrent = (RubikWidget)current;
	RubikWidget rrequest = (RubikWidget)request;
	RubikWidget rw = (RubikWidget)w;
*/
	Boolean redisplay = False;

	return redisplay;
}
#endif

static void
RubikDestroy(Widget w)
{
/*	RubikWidget rw = (RubikWidget)w;	*/
}

static void
RubikResize(Widget w)
{
	RubikWidget rw = (RubikWidget)w;
	XtWidgetProc resize;

	/* Call the superclass's resize, if needed */
	if ((resize = XtSuperclass(w)->core_class.resize))
		(*resize)(w);
	compute_cube_offset(rw);
}

static int
parsemotion(Widget w, String *params, Cardinal num_params, 
		int *dirp, int *sidep)
{
	static const String sidename[] = {
		"top", "bottom", "left", "right", "front", "back"
	};
	static const String dirname[] = {
		"ccw", "cw"
	};

	RubikWidget rw = (RubikWidget)w;
	int dir, side;

	if (num_params != 2) {
		XtAppWarning(XtWidgetToApplicationContext(w),
			"Turn/Roll: expected two parameters");
		return 0;
	}

	for (side = 0; side < XtNumber(sidename); side++)
		if (strcmp(sidename[side], params[0]) == 0)
			break;
	if (side >= XtNumber(sidename)) {
		XtAppWarning(XtWidgetToApplicationContext(w),
			"Turn/Roll: 1st param must be a side name");
		return 0;
	}

	for (dir = 0; dir < XtNumber(dirname); dir++)
		if (strcmp(dirname[dir], params[1]) == 0)
			break;
	if (dir >= XtNumber(dirname)) {
		XtAppWarning(XtWidgetToApplicationContext(w),
			"Turn/Roll: 2nd param must be 'ccw' or 'cw'");
		return 0;
	}

	rw->rubik.undo_face = sidename[side];
	rw->rubik.undo_dir = dirname[1-dir];

	*dirp = dir;
	*sidep = side;
	return 1;
}

/*
 * Action Implementation
 */
static void
Turn(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	RubikWidget rw = (RubikWidget)w;
	int dir, side;

	if (parsemotion(w, params, *num_params, &dir, &side)) {
		cube_turn(side, dir);
		rw->rubik.undo_action = "Turn";
		drawcube(w, 1);
	} else
		rw->rubik.undo_action = NULL;
}

static void
Roll(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	RubikWidget rw = (RubikWidget)w;
	int dir, side;

	if (parsemotion(w, params, *num_params, &dir, &side)) {
		cube_roll(side, dir);
		rw->rubik.undo_action = "Roll";
		drawcube(w, 1);
	} else
		rw->rubik.undo_action = NULL;
}

static void
Reset(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	RubikWidget rw = (RubikWidget)w;

	cube_reset();
	rw->rubik.undo_action = NULL;
	redisplay(w);
}

static void
Undo(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	String myparams[2];
	RubikWidget rw = (RubikWidget)w;

	if (rw->rubik.undo_action == NULL) {
		/* No action to reverse */
		return;
	}
	myparams[0] = rw->rubik.undo_face;
	myparams[1] = rw->rubik.undo_dir;
	XtCallActionProc(w, rw->rubik.undo_action, event, 
		myparams, XtNumber(myparams));    
}

/* Force an expose event over the window */
static void
redisplay(Widget w)
{
	XClearArea(XtDisplay(w), XtWindow(w), 0, 0, 
		w->core.width, w->core.height, True);
}

/*
 * Given a point in three space (0 <= x,y,z < 1), return a position inside
 * the window.
 */
static void
perspective(RubikWidget rw, double x, double y, double z, XPoint *p) {
	double xr, yr;

	perspectivef(rw, x, y, z, &xr, &yr);
	p->x = xr * rw->rubik.cube_scale + rw->rubik.cube_offset.x;
	p->y = yr * rw->rubik.cube_scale + rw->rubik.cube_offset.y;
}

/* Transform a coordinate in the 3x3x3 cube and project into 2-D space for viewing */
static void
perspectivef(RubikWidget rw, double x, double y, double z, 
		double *xp, double *yp)
{
	double tx,ty,tz;

	/* Shift cube to be origin centred */
	y -= 1.5;
	x -= 1.5;
	z -= 1.5;

	/* rotate 90 degrees about x axis */
	tx=x; ty=-z; tz=y;	

	/* rotate 45 degrees about y axis */
	x = M_SQRT1_2*tx + -M_SQRT1_2*tz;
	y = ty;
	z = M_SQRT1_2*tx + M_SQRT1_2*tz;

	/* rotate -45 degrees about the x axis */
	tx=x;ty=y;tz=z;
/*
	x = tx;
	y = M_SQRT1_2*ty + M_SQRT1_2*tz;
	z = -M_SQRT1_2*ty + M_SQRT1_2*tz;
*/

#define A ((-22.5)*M_PI/180)
#define COSA /*cos(A)=*/ 0.92387953251128674
#define SINA /*sin(A)=*/ (-0.38268343236508978)
	x = tx;
	y = COSA*ty - SINA*tz;
	z = SINA*ty + COSA*tz;

	/* move object back by f units */
	z += rw->rubik.focal_dist;

	/* Apply standard perspective transformation */
	*xp = x * rw->rubik.focal_dist / z;
	*yp = -y * rw->rubik.focal_dist / z;
}

/*
 * Compute the scale and offset for persective() so as to make the 
 * transfromd cube fit nicely into the widget frame.
 */
static void
compute_cube_offset(RubikWidget rw)
{

	static const double PAD = 0.9;	/* give 10% space around the cube */
	struct {
		double x, y;
	} p[6], tl, br;
	double w, h;
	int i;

	/* Compute the furthest 6 points of the cube */
	perspectivef(rw, 3, 3, 0, &p[0].x, &p[0].y);
	perspectivef(rw, 3, 0, 0, &p[1].x, &p[1].y);
	perspectivef(rw, 3, 0, 3, &p[2].x, &p[2].y);
	perspectivef(rw, 0, 0, 3, &p[3].x, &p[3].y);
	perspectivef(rw, 0, 3, 3, &p[4].x, &p[4].y);
	perspectivef(rw, 0, 3, 0, &p[5].x, &p[5].y);
	perspectivef(rw, 0, 0, 0, &tl.x, &tl.y);
	memcpy(&br, &tl, sizeof br);

	for (i = 0; i < 6; i++) {
		if (tl.x > p[i].x)
			tl.x = p[i].x;
		if (tl.y > p[i].y)
			tl.y = p[i].y;
		if (br.x < p[i].x)
			br.x = p[i].x;
		if (br.y < p[i].y)
			br.y = p[i].y;
	}

	w = br.x - tl.x;
	h = br.y - tl.y;

	if (((double)h * (PAD*rw->core.width) / w) < (PAD*rw->core.height))
		rw->rubik.cube_scale = (PAD * rw->core.width) / w;
	else
		rw->rubik.cube_scale = (PAD * rw->core.height) / h;

	h *= rw->rubik.cube_scale;
	w *= rw->rubik.cube_scale;
	tl.x *= rw->rubik.cube_scale;
	tl.y *= rw->rubik.cube_scale;

	rw->rubik.cube_offset.x = rw->core.width / 2 - (tl.x + w / 2);
	rw->rubik.cube_offset.y = rw->core.height / 2 - (tl.y + h / 2);
}
