/* $Id: Rubik.h,v 1.2 2003/01/25 14:31:55 d Exp $ */
/* David Leonard, 2003. Public domain. */

#ifndef _Rubik_h_
#define _Rubik_h_

/*
 A Rubiks cube display

 Class:		rubikWidgetClass
 Class Name:	Rubik
 Superclass:	widget

 Resources added by the Rubik widget class

 Name		Class		RepType		Default value
 ----		-----		-------		-------------
 top		Top		Pixel		XtDefaultBackground
 bottom		Bottom		Pixel		XtDefaultBackground
 right		Right		Pixel		XtDefaultBackground
 left		Left		Pixel		XtDefaultBackground
 front		Front		Pixel		XtDefaultBackground
 back		Back		Pixel		XtDefaultBackground
 focalDist	FocalDist	Float		30.0
 elevation	Elevation	Float		-10.0
 edgeThickness	EdgeThickness	Dimension	3

 The cube is drawn with faces colours determined by the 'top', 'bottom',
 'right', 'left', 'front', and 'back' resources. The cube is drawn using
 perspective projection from a distance of 'focalDist', after translating 
 the 3x3 cube a distance given by the 'elevation' resource. The cube's
 faces are drawn first, and the edge is drawn using the widget's foreground
 colour with a thickness determined by the 'edgeThickness' resource.

 The widget understands the following actions:

   Turn(face, direction)
	- turn the given face (one of 'top', 'bottom', etc.) in the direction
	  given relative (one of 'cw' or 'ccw' - meaning clockwise and
	  counter-clockwise, respectively).
   Roll(face, direction)
	- roll the cube about the face in the direction given.
	  Some rolls are equivalent, e.g. Roll(top,cw) == Roll(bottom,ccw).
   Reset()
	- set the cube to its solved state

*/

/*#define XtNtop	"top" */
#define XtNbottom	"bottom"
#define XtNright	"right"
#define XtNleft		"left"
#define XtNfront	"front"
#define XtNback		"back"
#define XtNfocalDist	"focalDist"
#define XtNelevation	"elevation"
#define XtNedgeThickness "edgeThickness"

#define XtCTop		"Top"
#define XtCBottom	"Bottom"
#define XtCRight	"Right"
#define XtCLeft		"Left"
#define XtCFront	"Front"
#define XtCBack		"Back"
#define XtCFocalDist	"FocalDist"
#define XtCElevation	"Elevation"
#define XtCEdgeThickness "EdgeThickness"

typedef struct _RubikClassRec *RubikWidgetClass;
typedef struct _RubikRec *RubikWidget;

extern WidgetClass rubikWidgetClass;

#endif /* _Rubik_h */
