/* $Id: xrubik.c,v 1.3 2003/01/25 16:24:28 d Exp $ */
/* David Leonard, 2003. Public domain. */

/*
 * Set up a top-level shell and some buttons to drive the Rubik widget
 */

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xmu/Editres.h>

#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include "Rubik.h"

#include <X11/Xaw/Cardinals.h>

static void	Quit(Widget, XEvent *, String *, Cardinal *);
static void	Turn(Widget, XEvent *, String *, Cardinal *);
static void	Roll(Widget, XEvent *, String *, Cardinal *);
static void	Undo(Widget, XEvent *, String *, Cardinal *);
static void	Reset(Widget, XEvent *, String *, Cardinal *);

static Widget	findrubik(Widget);

String fallback_resources[] = {
	"*rubik.width: 100",
	"*rubik.height: 100",
	NULL
};

XtActionsRec actions[] = {
	{ "Quit",	Quit },
	{ "Reset",	Reset },
	{ "Turn",	Turn },
	{ "Roll",	Roll },
	{ "Undo",	Undo },
};

/* Quit the application */
static void
Quit(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	XtAppSetExitFlag(XtWidgetToApplicationContext(w));
	exit(0);
}

/* Fund the Rubik widget so we can route actions to it */
static Widget 
findrubik(Widget w)
{
	Widget top, rubik = NULL;

	for (top = w; top && !XtIsTopLevelShell(top); top = XtParent(top))
		;
	if (top) 
		rubik = XtNameToWidget(top, "*rubik");
	if (!rubik)
		XtAppWarning(XtWidgetToApplicationContext(w),
			"Cannot find rubik widget~");
	return rubik;
}

/* Send Turn actions to *rubik */
static void
Turn(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	Widget rubik = findrubik(w);
	if (rubik)
		XtCallActionProc(rubik, "Turn", event, params, *num_params);	
}

/* Send Roll actions to *rubik */
static void
Roll(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	Widget rubik = findrubik(w);
	if (rubik)
		XtCallActionProc(rubik, "Roll", event, params, *num_params);	
}

/* Send Undo actions to *rubik */
static void
Undo(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	Widget rubik = findrubik(w);
	if (rubik)
		XtCallActionProc(rubik, "Undo", event, params, *num_params);	
}

/* Send Reset actions to *rubik */
static void
Reset(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	Widget rubik = findrubik(w);
	if (rubik)
		XtCallActionProc(rubik, "Reset", event, params, *num_params);	
}

int
main(int argc, char *argv[])
{
	XtAppContext app_con;
	Widget toplevel, paned, box, box2;
	extern void FixXaw(void);
	int i;

	FixXaw();

	toplevel = XtAppInitialize(&app_con, "Xrubik", NULL, ZERO,
		&argc, argv, fallback_resources, NULL, ZERO);

	XtAppAddActions(app_con, actions, XtNumber(actions));

	/*
	 * Create a panel and some buttons.
	 * (This relies heavily on the resource file being present!)
	 */
	paned = XtCreateManagedWidget("paned", panedWidgetClass, toplevel, 
		NULL, ZERO);
	XtCreateManagedWidget("rubik", rubikWidgetClass, paned, 
		NULL, ZERO);
	box = XtCreateManagedWidget("box", boxWidgetClass, paned, 
		NULL, ZERO);
	for (i = 1; i <= 12; i++) {
		char turnname[] = "turn999";
		snprintf(turnname, sizeof turnname, "turn%d", i);
		XtCreateManagedWidget(turnname,
			commandWidgetClass, box, NULL, ZERO);
	}
	for (i = 1; i <= 6; i++) {
		char rollname[] = "roll999";
		snprintf(rollname, sizeof rollname, "roll%d", i);
		XtCreateManagedWidget(rollname,
			commandWidgetClass, box, NULL, ZERO);
	}

	box2 = XtCreateManagedWidget("box2", boxWidgetClass, paned, 
		NULL, ZERO);
	XtCreateManagedWidget("reset", commandWidgetClass, box2, 
		NULL, ZERO);
	XtCreateManagedWidget("quit", commandWidgetClass, box2, 
		NULL, ZERO);
	XtCreateManagedWidget("label", labelWidgetClass, box2, 
		NULL, ZERO);

	/*
	 * Start the application main loop
	 */
	XtRealizeWidget(toplevel);
	XtAddEventHandler(toplevel, 0, True, _XEditResCheckMessages, NULL);
	XtAppMainLoop(app_con);

	exit(0);
}
